import typer

from flay.common.logging import enable_debug_logging
from .bundle import debug_bundle_app

debug_app = typer.Typer()
debug_app.add_typer(debug_bundle_app, name="bundle")


@debug_app.callback()
def debug_app_main() -> None:
    enable_debug_logging()



from __future__ import annotations
from importlib.util import resolve_name
from pathlib import Path
from libcst import ImportStar, Module, parse_module, ImportFrom
from libcst.helpers import get_full_name_for_node_or_raise
from flay.common.exc import ParsingError


def file_to_node(file: Path | str) -> Module | None:
    if isinstance(file, str):
        file = Path(file)
    if file.match("*.so"):
        return None
    try:
        file_content = file.read_text()
        return parse_module(file_content)
    except UnicodeDecodeError as e:
        raise ParsingError(
            f"Could not open {file} because it's not encoded with {e.encoding}"
        )


def get_import_from_absolute_module_spec(
    node: ImportFrom, parent_package: str | None = None
) -> list[str]:
    module_node = node.module
    if module_node is None and not node.relative:
        raise ParsingError(
            f"No absolute module spec could be found for {node}, {parent_package}"
        )
    if module_node:
        return [
            resolve_name(
                "." * len(node.relative) + get_full_name_for_node_or_raise(module_node),
                parent_package,
            )
        ]
    elif parent_package is not None and len(node.relative) == 1:
        res = [parent_package]
        if not isinstance(node.names, ImportStar):
            for name in node.names:
                res.append(f"{parent_package}.{name.evaluated_name}")
        return res

    raise ParsingError(
        f"Don't know how to construct absolute module spec for node {node}"
    )



from .app import app
import typing as t
from flay.common.logging import logfile_path_context
import logging

__all__ = ("cli",)
log = logging.getLogger(__name__)


def cli() -> t.Any:
    try:
        return app()
    except Exception as e:
        log.error(
            f"Unexpected error occurred. Detailed log can be found at {logfile_path_context.get()}",
            exc_info=True,
        )
        raise e


if __name__ == "__main__":
    cli()



from __future__ import annotations
from libcst import (
    Import,
    BatchableCSTVisitor,
    ImportFrom,
    visit_batched,
    Module,
)
from pathlib import Path
import typing as t

from flay.common.libcst import file_to_node, get_import_from_absolute_module_spec
from ..common.module_spec import find_module_path, get_parent_package
from stdlib_list import in_stdlib
import logging

log = logging.getLogger(__name__)


ModuleCollection = dict[tuple[str, Path], t.Optional[Module]]


class FileCollector(BatchableCSTVisitor):
    def __init__(
        self,
        package: str | None = None,
        already_collected: ModuleCollection | None = None,
    ) -> None:
        self._package = package
        self.collected_files: ModuleCollection = already_collected or {}

        super().__init__()

    def _process_module(self, module_spec: str) -> None:
        if in_stdlib(module_spec):
            return
        file_path = find_module_path(module_spec)

        if file_path is None or file_path.origin is None:
            log.warning(f"Don't know how to import {module_spec}. Skipping...")
            return
        file_path_origin = Path(file_path.origin)
        if (file_path.name, file_path_origin) in self.collected_files:
            return

        parsed_module_content = file_to_node(file_path.origin)
        self.collected_files[(file_path.name, file_path_origin)] = parsed_module_content
        log.debug(
            f"Collected {file_path.origin} from package {self._package} as {file_path.name}"
        )
        if parsed_module_content is None:
            return

        sub_collector = FileCollector(
            package=get_parent_package(file_path.name)
            if not file_path.origin.endswith("__init__.py")
            else file_path.name,
            already_collected=self.collected_files,
        )
        visit_batched(parsed_module_content, [sub_collector])
        self.collected_files.update(sub_collector.collected_files)

    def visit_Import(self, node: Import) -> None:
        for name in node.names:
            self._process_module(name.evaluated_name)

    def visit_ImportFrom(self, node: ImportFrom) -> None:
        for absolute_module in get_import_from_absolute_module_spec(
            node, self._package
        ):
            self._process_module(absolute_module)



from __future__ import annotations
from importlib.machinery import ModuleSpec
import sys
import typing as t
from pathlib import Path
from .exc import FlayFileNotFoundError
import os

from functools import cache


def get_top_level_package(module_spec: str) -> str:
    if "." not in module_spec:
        return module_spec
    return module_spec.split(".", 1)[0]


def get_parent_package(module_spec: str) -> str:
    if "." not in module_spec:
        return module_spec
    return ".".join(module_spec.split(".")[:-1])


VALID_FILE_EXTENSIONS = {"py", "so"}


def _path_can_lead_to_module_spec(
    path: str, file_name: str, module_spec_parts: list[str]
) -> bool:
    for i in range(len(module_spec_parts)):
        search_value = os.path.sep.join(module_spec_parts[:i])
        if f"{path}/{file_name}".endswith(search_value) or path.endswith(search_value):
            return True
    return False


@cache
def _lookup_paths_for_module_spec(module_spec: str) -> list[str] | None:
    if "." not in module_spec:
        return None
    module_spec_parts = module_spec.split(".")
    top_level = module_spec_parts[0]
    paths: set[str] = set()
    for sys_path in sys.path:
        if os.path.isdir(sys_path) and top_level in os.listdir(sys_path):
            top_level_path = f"{sys_path}/{top_level}"
            if not os.path.isdir(top_level_path) or "__init__.py" not in os.listdir(
                top_level_path
            ):
                continue
            paths.add(top_level_path)
            for path, dirs, files in os.walk(top_level_path):
                for file in files:
                    file_name, file_extension = file.rsplit(".", 1)

                    if _path_can_lead_to_module_spec(
                        path, file_name, module_spec_parts
                    ) and (
                        file_extension in VALID_FILE_EXTENSIONS
                        or (
                            not path.endswith("__pycache__") and file_extension == "pyc"
                        )
                    ):
                        paths.add(path)
    return list(paths)


def find_module_path(
    module_spec: str,
) -> ModuleSpec | None:
    lookup_paths = _lookup_paths_for_module_spec(module_spec)
    for finder in sys.meta_path:
        result = finder.find_spec(module_spec, lookup_paths)
        if result is not None:
            return result

    return None


def find_all_files_in_module_spec(module_spec: str) -> t.Generator[Path, t.Any, None]:
    found_path = find_module_path(module_spec)
    if found_path is None or found_path.origin is None:
        raise FlayFileNotFoundError(
            f"Could not find file for module spec '{module_spec}'"
        )
    module_init_file_path = Path(found_path.origin)

    module_folder_path = module_init_file_path.parent
    for file in os.listdir(str(module_folder_path)):
        if file.endswith(".py"):
            yield module_folder_path / file



from __future__ import annotations
from libcst import visit_batched
from stdlib_list import in_stdlib
from flay.bundle.collector import FileCollector
from flay.common.libcst import file_to_node
from flay.common.logging import log_cst_code
from flay.common.module_spec import find_all_files_in_module_spec, get_top_level_package
from libcst import CSTTransformer, Import, ImportFrom
from libcst import Attribute, Name
from pathlib import Path
import typing as t
from libcst.helpers import get_absolute_module_for_import_or_raise
import logging
import os.path
from libcst.metadata import ScopeProvider
from libcst import MetadataWrapper
import shutil

log = logging.getLogger(__name__)


class ImportsTransformer(CSTTransformer):
    METADATA_DEPENDENCIES = (ScopeProvider,)

    def __init__(
        self, top_level_package: str, vendor_module_name: str = "_vendor"
    ) -> None:
        self.top_level_package = top_level_package
        self.vendor_module_name = vendor_module_name
        self._affected_attributes: list[Attribute] = []
        self._affected_names: list[Name] = []
        super().__init__()

    def _prepend_vendor(self, node: Attribute | Name) -> Attribute:
        if isinstance(node, Name):
            new_name = Attribute(
                value=Attribute(
                    value=Name(self.top_level_package),
                    attr=Name(self.vendor_module_name),
                ),
                attr=node,
            )
        else:
            deepest_attribute = node
            while not isinstance(deepest_attribute.value, Name):
                deepest_attribute = t.cast(Attribute, deepest_attribute.value)

            new_name = node.with_deep_changes(
                deepest_attribute,
                value=Attribute(
                    value=Attribute(
                        value=Name(self.top_level_package),
                        attr=Name(self.vendor_module_name),
                    ),
                    attr=deepest_attribute.value,
                ),
            )

        return new_name

    def _prepend_vendor_for_import(
        self,
        node: Attribute | Name,
        module_spec: str,
        references_need_update: bool = False,
    ) -> Attribute | Name:
        if module_spec.startswith(self.top_level_package) or in_stdlib(module_spec):
            return node
        if references_need_update:
            if isinstance(node, Name):
                self._affected_names.append(node)
            else:
                self._affected_attributes.append(node)
        return self._prepend_vendor(node)

    def leave_Import(self, original_node: Import, updated_node: Import) -> Import:
        new_node = updated_node.with_changes(
            names=[
                name.with_changes(
                    name=self._prepend_vendor_for_import(
                        name.name,
                        name.evaluated_name,
                        references_need_update=name.asname is None,
                    )
                )
                for name in updated_node.names
            ]
        )
        log.debug(
            "Transformed '%s' to '%s'",
            log_cst_code(original_node),
            log_cst_code(new_node),
        )
        return new_node

    def leave_ImportFrom(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> ImportFrom:
        if updated_node.module and not updated_node.relative:
            module_spec = get_absolute_module_for_import_or_raise(None, updated_node)
            new_node = updated_node.with_changes(
                module=self._prepend_vendor_for_import(updated_node.module, module_spec)
            )

            log.debug(
                "Transformed '%s' to '%s'",
                log_cst_code(original_node),
                log_cst_code(new_node),
            )
            return new_node
        return updated_node

    def leave_Name(self, original_node: Name, updated_node: Name) -> Name | Attribute:
        # TODO: we need to make that we are inside the scope of the original import
        for maybe_name in self._affected_names:
            if maybe_name.deep_equals(updated_node):
                new_node = self._prepend_vendor(updated_node)
                log.debug(
                    "Transformed Name '%s' to '%s'",
                    log_cst_code(updated_node),
                    log_cst_code(new_node),
                )
                return new_node
        return updated_node

    def leave_Attribute(
        self, original_node: Attribute, updated_node: Attribute
    ) -> Attribute:
        # TODO: we need to make that we are inside the scope of the original import
        for maybe_attribute in self._affected_attributes:
            if maybe_attribute.deep_equals(updated_node):
                new_node = self._prepend_vendor(updated_node)
                log.debug(
                    "Transformed Attribute '%s' to '%s'",
                    log_cst_code(updated_node),
                    log_cst_code(new_node),
                )
                return new_node
        return updated_node


def bundle_package(
    module_spec: str, destination_path: Path, vendor_module_name: str = "_vendor"
) -> None:
    collector = FileCollector(package=module_spec)
    for path in find_all_files_in_module_spec(module_spec):
        module = file_to_node(path)
        if module is not None:
            visit_batched(module, [collector])

    files = collector.collected_files
    top_level_package = get_top_level_package(module_spec)
    imports_transformer = ImportsTransformer(
        top_level_package=top_level_package,
        vendor_module_name=vendor_module_name,
    )
    for key, node in files.items():
        if node:
            files[key] = MetadataWrapper(node).visit(imports_transformer)

    vendor_path = destination_path / top_level_package / vendor_module_name

    gitignore = destination_path / ".gitignore"
    if not gitignore.exists():
        gitignore.parent.mkdir(parents=True, exist_ok=True)
        gitignore.write_text("*")

    for (found_module, found_path), module_node in files.items():
        module_path_part = Path(os.path.sep.join(found_module.split(".")))
        is_external = get_top_level_package(found_module) != top_level_package

        if found_path.match(f"*/{module_path_part}/__init__.py"):
            if is_external:
                target_file = vendor_path / module_path_part / "__init__.py"
            else:
                target_file = destination_path / module_path_part / "__init__.py"
        elif is_external:
            target_file = vendor_path / module_path_part.parent / found_path.name
        else:
            target_file = destination_path / module_path_part.parent / found_path.name

        target_dir = target_file.parent
        if not target_dir.exists():
            target_dir.mkdir(parents=True)
        if module_node:
            target_file.write_text(module_node.code)
        else:
            shutil.copy2(str(found_path), str(target_file))



from __future__ import annotations
import logging
from libcst import CSTNode, Module
import platformdirs
import tempfile
import contextvars
import typing as t

logfile_path_context: contextvars.ContextVar[str] = contextvars.ContextVar(
    "_flay_logfile"
)


def get_flay_logger() -> logging.Logger:
    return logging.getLogger("flay")


def setup_logger(command: str) -> None:
    flay_logger = get_flay_logger()
    stream_handler = logging.StreamHandler()
    logging_root_dir = platformdirs.user_log_dir("flay", ensure_exists=True)
    logging_file_path = tempfile.mktemp(".log", f"flay-{command}", logging_root_dir)
    logfile_path_context.set(logging_file_path)
    file_handler = logging.FileHandler(logging_file_path)
    flay_logger.addHandler(stream_handler)
    flay_logger.addHandler(file_handler)


def enable_debug_logging() -> None:
    flay_logger = get_flay_logger()
    flay_logger.setLevel(logging.DEBUG)


class _Serializable(t.Protocol):
    def __str__(self) -> str: ...


class LazyStr:
    def __init__(self, factory: t.Callable[[], str | _Serializable]):
        self.factory = factory

    def get_string(self) -> str:
        resolved = self.factory()
        if isinstance(resolved, str):
            return resolved
        return str(resolved)

    def __repr__(self) -> str:
        return f"<LazyStr resolved_value='{self.get_string()}' >"

    def __str__(self) -> str:
        return self.get_string()


def log_cst_code(node: CSTNode) -> LazyStr:
    """
    A helper method for when the code of a CST node needs to be logged.
    It returns a LazyStr which only evaluates the needed code when serialized.
    """

    def gen_str() -> str:
        fake_module = Module([node])  # type: ignore
        return fake_module.code_for_node(node)

    return LazyStr(gen_str)



import typer

from flay.common.logging import setup_logger
from .debug import debug_app

app = typer.Typer(pretty_exceptions_enable=False)

app.add_typer(debug_app, name="debug")


@app.callback()
def app_main() -> None:
    setup_logger("")



import typer
from flay.bundle.collector import FileCollector
from libcst import visit_batched
import typing as t

from flay.bundle.package import bundle_package
from ...common.module_spec import find_all_files_in_module_spec
from ...common.libcst import file_to_node
import logging
from pathlib import Path
import shutil

debug_bundle_app = typer.Typer()

log = logging.getLogger(__name__)


@debug_bundle_app.command("collector")
def bundle_collector(
    module_spec: t.Annotated[str, typer.Argument(help="A module spec to test")],
) -> None:
    collector = FileCollector(package=module_spec)
    for path in find_all_files_in_module_spec(module_spec):
        log.debug(f"Found: {path}")
        module_node = file_to_node(path)
        if module_node:
            visit_batched(module_node, [collector])

    typer.echo({str(k): type(v) for k, v in collector.collected_files.items()})


@debug_bundle_app.command("bundle_package")
def debug_bundle_package(
    module_spec: t.Annotated[str, typer.Argument(help="A module spec to test")],
    dest_path: t.Annotated[
        Path,
        typer.Argument(
            default_factory=lambda: Path("./debug_bundle"),
            help="Destination path for the completed bundle",
            resolve_path=True,
        ),
    ],
) -> None:
    if dest_path.exists():
        shutil.rmtree(str(dest_path))
    bundle_package(module_spec=module_spec, destination_path=dest_path)
